//g++ -o matrix matrix.cpp -I c:\GnuWin32\include  -L c:\GnuWin32\lib -lgsl -lgslcblas -lm

/*
 * K-Means Code
 *
 * Emre Tasci / Nov 2007, TUDelft
 *
 * e.tasci@tudelft.nl
 *
 * Generates given number of data randomly,
 * then initializes given number of means
 * and associates the data to the means using
 * the criteria that is defined in the 
 * distance function located in km.h
 *
 * Required External Libraries:
 * gsl
 * TCLAP - Since it is used only for argument
 * 	parsing, may be safely removed.
 *
 * Possible Improvements:
 * - Option to read the "World" from file as well as 
 * generating it.
 * - Improvement of the Means' initial positions.
 *
 * Input Parameters:
 *
 * USAGE:
 * 
 *    kmeans  -n <int> -k <int> [-s <unsigned long int>] [-o <output file>]
 *            [-m <int>] [-q] [--] [--version] [-h]
 * 
 * 
 * Where:
 * 
 *    -n <int>,  --numberofelements <int>
 *      (required)  total number of elements (required)
 * 
 *    -k <int>,  --numberofmeans <int>
 *      (required)  total number of means (required)
 * 
 *    -s <unsigned long int>,  --seed <unsigned long int>
 *      seed for random number generator (default: timer)
 * 
 *    -o <output file>,  --output <output file>
 *      output file basename (default: "temp")
 * 
 *    -m <int>,  --maxcounter <int>
 *      maximum number of steps allowed (default: 30)
 * 
 *    -q,  --quiet
 *      no output to screen is generated
 * 
 *    --,  --ignore_rest
 *      Ignores the rest of the labeled arguments following this flag.
 * 
 *    --version
 *      Displays version information and exits.
 * 
 *    -h,  --help
 *      Displays usage information and exits.
 * 
 * Output Files:
 * <project-name>.means.txt : Coordinates of the means evolution in the iterations
 * <project-name>.meansfinal.txt : Final coordinates of the means
 * <project-name>.report : A summary of the process
 * <project-name>.world.dat : The "world" matrice in Mathematica Write format
 * <project-name>.world.txt : Coordinates of the items (ie, the "world" matrice in XYZ format)
 * <project-name>-mean-#-resp.txt : Coordinates of the items belonging to the #. mean
 */

#include <stdio.h>
#include "matrixop.h"
#include "km.h"

//TCLAP Section.........
#include <exception>
#include "tclap/CmdLine.h"
using namespace TCLAP;
int _dn;
int _dk;
bool _quiet;
unsigned int _dmaxcounter;
unsigned long int _dseed;
string _dprojectname;
//TCLAP Section.........


//Global Variables:
unsigned long int seed = 0;
int dN = 5; // number of items
int dK = 4; // number of means
string dProjectName(""); // name of the project to be used in file operations
unsigned int dMaxCounter = 10;
#include "matrixop.cpp"
#include "km.cpp"

void parseOptions(int argc, char** argv)
{
	//Using TCLAP for parsing the command line options
	//(http://tclap.sourceforge.net/) by Michael E. Smoot
	
	try {

	CmdLine cmd("K-Means by Emre S. Tasci, <e.tasci@tudelft.nl> 2007", ' ', "1.00" );

	//
	// Define arguments
	//

	SwitchArg quiet("q","quiet","no output to screen is generated", false);
	cmd.add(quiet );
	
	ValueArg<unsigned int> dmaxcounter("m", "maxcounter", "maximum number of steps allowed (default: 30)", false, 30, "int");
	cmd.add( dmaxcounter );

	ValueArg<string> dprojectname("o","output","output file basename (default: \"temp\")",false,"temp","output file");
	cmd.add( dprojectname );

	ValueArg<unsigned long int> dseed("s", "seed", "seed for random number generator (default: timer)", false, 0, "unsigned long int");
	cmd.add( dseed );

	ValueArg<int> dk("k", "numberofmeans", "total number of means (required)", true, -240, "int");
	cmd.add( dk );

	ValueArg<int> dn("n", "numberofelements", "total number of elements (required)", true, -240, "int");
	cmd.add( dn );


	//
	// Parse the command line.
	//
	cmd.parse(argc,argv);

	//
	// Set variables
	//
	_dn   		= dn.getValue();
	_dk  		= dk.getValue();
	_dseed        	= dseed.getValue();
	_dprojectname	= dprojectname.getValue();
	_dmaxcounter 	= dmaxcounter.getValue();
	_quiet       	= quiet.getValue();

	} catch ( ArgException& e )
	{ cout << "ERROR: " << e.error() << " " << e.argId() << endl; exit(5); }
}

int writemeans2file(string baseFileName,kmean ** means, gsl_matrix *m)
{
	FILE *pFile;
	FILE *pFileMeansFinal;
	pFile = fopen(extension_changer_adderEST(baseFileName,"report.txt").c_str(),"a");
	pFileMeansFinal = fopen(extension_changer_adderEST(baseFileName,"meansfinal.txt").c_str(),"w");

	//Final Coordinates:
	fprintf(pFile,"Final Coordinates of the Means\n");
	for(int i=0;i<dK;i++)
	{
		fprintf(pFile,"Mean %d : %10.5f\t%10.5f\t%10.5f\n",i,means[i]->XX(),means[i]->YY(),means[i]->ZZ());
		fprintf(pFileMeansFinal,"%10.5f\t%10.5f\t%10.5f\n",means[i]->XX(),means[i]->YY(),means[i]->ZZ());
	}

	//Responsibilities
	fprintf(pFile,"\nResponsibilities for %d data points\n",dN);
	double weight = 0;
	for(int i=0;i<dK;i++)
	{
		weight = (means[i]->RTOT()+0.0) / (dN+0.0) * 100.0;
		fprintf(pFile,"Mean %d has %d responsibilities (%5.1f%% weight)\n",i,means[i]->RTOT(),weight);
	}
	for(int i=0;i<dK;i++)
	{
	  	char buffer [33];
		FILE *pFile1;
  		itoa (i,buffer,10);
		string baseFileName2("");
		baseFileName2 = baseFileName+"-mean-"+buffer+"-resp";
		pFile1 = fopen(extension_changer_adderEST(baseFileName2,"txt").c_str(),"w");
		fprintf(pFile,"\nResponsibilities for Mean %d:\n",i);
		int j,counterj;
		for(j=0,counterj=0;j<dN;j++)
			if(gsl_vector_get(means[i]->RRV(),j))
			{
				if(counterj<20)fprintf(pFile,"%d ",j);
				else {fprintf(pFile,"%d \n",j); counterj=0;}
				fprintf(pFile1,"% 10.5f\t% 10.5f\t% 10.5f\n", gsl_matrix_get(m,j,1),gsl_matrix_get(m,j,2),gsl_matrix_get(m,j,3));

			}
		fprintf(pFile,"\n");
		fclose(pFile1);
	}
	


	fclose(pFile);
	fclose(pFileMeansFinal);
	return 0;
}

int main (int argc, char* argv[])
{
	//TCLAP - Parsing the command line options
	parseOptions(argc,argv);
	
	if (_dk>0) dK 		= _dk;
	if (_dn>0) dN 		= _dn;
	if (_dseed>0) seed 	= _dseed;
	dMaxCounter 		= _dmaxcounter;
	dProjectName 		= _dprojectname;
	

	int i = 0;//iteration variable
	int counter = 1;
	double xx,yy,zz;//temporary mean positions
	bool f_maxed = false;//a flag to keep track of maxing the dMaxCounter

	M_InitializeEST();//initializing the matrixop.h

	gsl_matrix * m = gsl_matrix_alloc (dN, 4);//the world
	M_RandomizerEST(dN,m,-100.0,100.0);//the world, randomized
	M_Write2FileEST(dProjectName,m,dN,false);//Saving the world (for further reference)

	//M_PrintEST(m,dN);

	gsl_matrix * distances = gsl_matrix_alloc (dK, dN); //the distance matrix
	gsl_matrix_set_all(distances,100.0);//Initializing the distance matrix with "high" values

	//M_PrintEST(distances,dK,4,false);
	
	// we will store the mean evolutions in this file
	FILE *pFile4;
	pFile4 = fopen(extension_changer_adderEST(dProjectName,"means.txt").c_str(),"w"); 
	
	//Initializing the means:
	kmean **means;
	means = new kmean * [dK];
	
	gsl_matrix *prevMeanPositions 	= gsl_matrix_calloc (dK,3);//Stores the previous step's mean positions
						 		   //so that we can check for continuation.
	gsl_matrix *currMeanPositions 	= gsl_matrix_calloc (dK,3);
	
	for (i=0;i<dK;i++)
	{
		xx = 3.0;
		yy = 3.0+i;
		zz = 3.0;
		means[i] = new kmean(i,xx,yy,zz,dN); // The initial positions will be refined later XXX
		gsl_matrix_set(prevMeanPositions,i,0,xx);
		gsl_matrix_set(prevMeanPositions,i,1,yy);
		gsl_matrix_set(prevMeanPositions,i,2,zz);
	}
	

	if(!_quiet)
	{
		printf("Initial Coordinates of the Means:\n");
		for (i=0;i<dK;i++)
			printf("Coordinates of mean %d:\t%s",i,means[i]->whereis());
	}
		for (i=0;i<dK;i++)
			fprintf(pFile4,"%s",means[i]->whereis());

analyze:
	{
		//Assigns the data to the nearest mean..

		if(!_quiet)printf("\n-----%d. STEP --------\n",counter);


		//Calculating the distances and assigning the data
		//to the nearest mean via the function: check_distances:
		for (i=0;i<dK;i++)
			means[i]->calculate_distances(m,dN,distances);
		
		check_distances(distances,dN,means);
	
		/* Info---------------------------------------------------
		M_PrintEST(distances,dK,4,false);
		for(i=0;i<dK;i++)
		{
			printf("Mean #%d responsibilities:\n",i);
			means[i]->RR();
			printf("\n");
		}
		Info ---------------------------------------------------*/
	}



move:
	{
		//moves the means wrt to the data's center of mass
		//they're responsible for..

		for(i=0; i<dK; i++)
		{
			means[i]->moveCoM(m);
			gsl_matrix_set(currMeanPositions,i,0,means[i]->XX());
			gsl_matrix_set(currMeanPositions,i,1,means[i]->YY());
			gsl_matrix_set(currMeanPositions,i,2,means[i]->YY());
		}
		if(!_quiet)
		{
			printf("\n");
			for (i=0;i<dK;i++)
				printf("Coordinates of mean %d:\t%s",i,means[i]->whereis());
		}
		for (i=0;i<dK;i++)
			fprintf(pFile4,"%s",means[i]->whereis());
	}
	
checkSufficiency:
	{
		//checks if any of the means has moved in this step.
		//If not, ends the program
		//
		//also checks if maximum number of steps has been
		//achieved. If so, ends the program

		//checking the counter is easier, so we will start
		//with it:
		counter++;
		if(counter>=dMaxCounter) {f_maxed = true;goto finalize;}

		gsl_matrix_sub(prevMeanPositions,currMeanPositions);
		if(gsl_matrix_isnull(prevMeanPositions)) goto finalize;
		
		gsl_matrix_memcpy(prevMeanPositions,currMeanPositions);
		goto analyze;
	}



	

	/*
	//Reading data in a file into a matrix:
	FILE * pFile2;
	pFile2 = fopen("d:\\lol1.dat","r");
	gsl_matrix * k = gsl_matrix_alloc (20,4);
	gsl_matrix_fscanf(pFile2,k);
	//gsl_matrix_fprintf(stdout,k,"%f");
	fclose(pFile2);	
	*/

finalize:
	{
		//Writing the properties for the means:
		writemeans2file(dProjectName,means,m);
		
		fclose(pFile4);


		//Tidying Up
		gsl_matrix_free (m);
		gsl_matrix_free (distances);
		gsl_matrix_free (prevMeanPositions);
		gsl_matrix_free (currMeanPositions);

		for(i=0; i<dK; i++)
			delete means[i];
	}

	if(!f_maxed)printf("\n\nSuccess in %d steps!\n",--counter);
	else printf("\n\nStopped in %d steps due to the maximum number of steps limit!\n",counter);

	return 0;
}
